//
//  tank
//
//  Import and Playlist Daemon for autoradio project
//
//
//  Copyright (C) 2017-2018 Christian Pointner <equinox@helsinki.at>
//
//  This file is part of tank.
//
//  tank is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  any later version.
//
//  tank is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with tank. If not, see <http://www.gnu.org/licenses/>.
//

package importer

import (
	"encoding/base64"
	"io"
	"sync/atomic"

	"golang.org/x/crypto/blake2b"
)

func (job *Job) prepareSource() {
	switch job.Source.Scheme {
	case SourceSchemeAttachment:
		// the source will be attached using AttachSource()
		return

		// TODO: implement other sources

	default:
		// simulate a 10 MB file...
		atomic.StoreUint32(&job.sourceSet, 1)
		job.source = newJobSourceNull(10 * 1024 * 1024)
		close(job.subC.sourceAttached)
	}
}

type copyResult struct {
	err  error
	hash string
}

func (job *Job) fetch() (interface{}, error) {
	job.Progress.set(StepFetching, 0)

	// wait until source is attached
	select {
	case <-job.ctx.Done():
		return nil, job.ctx.Err()
	case <-job.subC.sourceAttached:
	}
	// job.source is now initialized and points to a valid source
	// make sure a potentially connected source gets notified in any case
	defer job.source.Done(nil)

	conv, err := job.newFetchConverter()
	if err != nil {
		job.im.errLog.Printf("fetch(): creating fetch converter failed: %v", err)
		return nil, err
	}
	// from here on conv.Close() and conv.Wait() has to be called in any case to
	// reap potential child process zombies

	done := make(chan copyResult)
	go func() {
		hash, err := blake2b.New256(nil)
		if err != nil {
			panic("creating hash function failed: " + err.Error())
		}
		src := io.TeeReader(job.source, hash)
		written, err := io.Copy(&progressWriter{job, StepFetching, job.source.Len(), 0, conv}, src)
		if err != nil {
			done <- copyResult{err, ""}
			return
		}
		hashStr := "blake2b_256:" + base64.URLEncoding.EncodeToString(hash.Sum(nil))
		job.im.dbgLog.Printf("fetch(): done copying %d bytes from source (%s)", written, hashStr)
		_, err = job.im.store.UpdateFileSourceHash(job.Group, job.ID, hashStr)
		done <- copyResult{err, hashStr}
	}()

	var res copyResult
	select {
	case <-job.ctx.Done():
		conv.Close()
		go conv.Wait() // do the zombie reaping in seperate go routine since we are not interested in the result anyway
		err = job.ctx.Err()
		job.source.Done(&JobSourceResult{Err: err})
		return nil, err
	case res = <-done:
	}

	conv.Close()
	loudness, convLog, err := conv.Wait()
	job.im.dbgLog.Printf("fetch(): converter returned: %T, %v", loudness, err)
	if res.err != nil {
		err = res.err
	}

	for _, l := range convLog {
		job.im.dbgLog.Println(l.Line)
	}
	job.Progress.set(StepFetching, 1)
	job.source.Done(&JobSourceResult{err, res.hash, convLog})
	return loudness, err
}
