//
//  tank
//
//  Import and Playlist Daemon for autoradio project
//
//
//  Copyright (C) 2017-2018 Christian Pointner <equinox@helsinki.at>
//
//  This file is part of tank.
//
//  tank is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  any later version.
//
//  tank is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with tank. If not, see <http://www.gnu.org/licenses/>.
//

package importer

import (
	"context"
	"io/ioutil"
	"log"
	"net/url"
	"runtime"
	"sync"

	"gitlab.servus.at/autoradio/tank/store"
)

type Importer struct {
	conf     *Config
	infoLog  *log.Logger
	errLog   *log.Logger
	dbgLog   *log.Logger
	store    *store.Store
	wgWorker sync.WaitGroup
	work     chan *Job
	jobs     *jobInventory
}

func (im *Importer) ListJobs(group string) (Jobs, error) {
	return im.jobs.ListJobs(group), nil // for now error is always nil but this might change later
}

func (im *Importer) CreateJob(group string, id uint64, src url.URL, user, refID string) (*Job, error) {
	// TODO: update this list once we implemented other sources
	switch src.Scheme {
	case SourceSchemeAttachment:
		//	case SourceSchemeFlowJS:
	default:
		return nil, ErrSourceNotSupported
	}

	file, err := im.store.GetFile(group, id)
	if err != nil {
		return nil, err
	}
	if file.Source.Import.State != store.ImportNew {
		return nil, ErrFileNotNew
	}

	return im.jobs.InsertOrGetJob(group, id, newJob(im, group, id, src, user, refID)), nil
}

func (im *Importer) GetJob(group string, id uint64) (*Job, error) {
	return im.jobs.GetJob(group, id)
}

// handle subscriptions to new jobs-of-group -> pass on to job inventory

func (im *Importer) deleteJob(group string, id uint64) error {
	return im.jobs.DeleteJob(group, id)
}

func (im *Importer) runWorker(idx int) {
	defer im.dbgLog.Printf("importer: worker(%d) has stopped", idx)

	im.dbgLog.Printf("importer: worker(%d) is running", idx)
	for job := range im.work {
		im.infoLog.Printf("importer: worker(%d) starting job(%s/%d) ...", idx, job.Group, job.ID)

		// job.run() will take care of cancelation and timeouts to make sure this function returns in time.
		// In case the job is already canceled we still need to run this to make sure all resources
		// allocated by the job are properly cleaned up and the job is removed from the inventory.
		if err := job.run(); err != nil {
			if err == context.Canceled {
				im.infoLog.Printf("importer: worker(%d) job(%s/%d) was canceled", idx, job.Group, job.ID)
			} else {
				im.errLog.Printf("importer: worker(%d) job(%s/%d) failed: %s", idx, job.Group, job.ID, err.Error())
			}
		} else {
			im.infoLog.Printf("importer: worker(%d) successfully completed job(%s/%d)", idx, job.Group, job.ID)
		}
	}
}

func NewImporter(conf Config, st *store.Store, infoLog, errLog, dbgLog *log.Logger) (im *Importer, err error) {
	if infoLog == nil {
		infoLog = log.New(ioutil.Discard, "", 0)
	}
	if errLog == nil {
		errLog = log.New(ioutil.Discard, "", 0)
	}
	if dbgLog == nil {
		dbgLog = log.New(ioutil.Discard, "", 0)
	}

	im = &Importer{conf: &conf, infoLog: infoLog, errLog: errLog, dbgLog: dbgLog, store: st}

	if im.conf.Workers <= 0 {
		im.conf.Workers = runtime.NumCPU()
	}
	if im.conf.Backlog == 0 {
		im.conf.Backlog = DefaultBacklog
	}

	im.work = make(chan *Job, im.conf.Backlog)
	for i := 0; i < im.conf.Workers; i = i + 1 {
		im.wgWorker.Add(1)
		go func(idx int) {
			defer im.wgWorker.Done()
			im.runWorker(idx)
		}(i)
	}
	im.jobs = newJobInventory(infoLog, errLog, dbgLog)

	infoLog.Printf("importer: started with %d worker", im.conf.Workers)
	return
}
